/*
 * i2c.c
 *
 *  Created on: 2013/12/21
 *      Author: d10130
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <unistd.h>
#include "header/debug.h"
#include "header/i2c.h"

/*---- I2Cマスターボードからの返答 ----*/
/* I2C通信が正常に開始された */
#define RESPONSE_SUCCESS           0x00
/* スレーブからの応答がない */
#define RESPONSE_SLAVE_NO_RESPONSE 0x01
/* I2Cマスターボードが正常に受信できなかった */
#define RESPONSE_RS232C_ERROR      0x02



/*---- 構造体の中身 ----*/
struct I2CBus_ {
	int fd;                    /* I2Cマスターボードのファイルディスクリプタ */
	char *name;                /* I2Cマスターボードのデバイスファイル名 */
	pthread_mutex_t mutex;     /* スレッド同期用mutex */
	I2CDevice* devices;        /* 循環リスト用のヘッダ */
	I2CBus* next;              /* 次のバスへのポインタ */
};
struct I2CDevice_ {
	I2CBus* bus;           /* デバイスが所属するバス */
	uint8_t address;       /* デバイスのアドレス */
	I2CDevice* next;       /* 次のデバイスへのポインタ */
	pthread_mutex_t mutex; /* スレッド同期用mutex */
};

/*---- グローバル変数 ----*/
/* 循環リスト用のヘッダの実体 */
I2CBus bussesHeader = {
		.next = &bussesHeader,
		.mutex = PTHREAD_MUTEX_INITIALIZER,
};
/* 循環リスト */
I2CBus* busses = &bussesHeader;

/*
 * 通信設定用関数
 * 引数: I2Cバスポインタ
 * 戻値: 正常時 0
 *       異常時 1
 */
static int i2cConfigTermios(I2CBus* bus) {
	int error;
	struct termios newTermios;
	/* 特に指定無し */
	newTermios.c_iflag = 0;
	/* 特に指定なし */
	newTermios.c_oflag = 0;
	/* データ長8bit, 受信許可, 制御線無し */
	newTermios.c_cflag = CS8 | CREAD | CLOCAL;
	/* 非カノニカルモード */
	newTermios.c_lflag = 0;
	/* 最小1Byteから即受信, 受信待機時間上限設定 */
	newTermios.c_cc[VMIN] = 0;
	newTermios.c_cc[VTIME] = 1; /* 0.1 sec 単位 */
	/* 通信速度設定 */
	cfsetispeed(&newTermios, I2C_BAUD_RATE);
	cfsetospeed(&newTermios, I2C_BAUD_RATE);
	/* 送受信バッファリセット */
	tcflush(bus->fd, TCIOFLUSH);
	/* 設定変更反映 */
	error = tcsetattr(bus->fd, TCSANOW, &newTermios);
	if (error != 0) {
		error = errno; /* errno退避 */
		ERROR_PRINT("tcsetattr: %s: %s\n",bus->name,strerror(errno));
		errno = error;
		return -1;
	}
	return 0;
}

/*
 * バスのオープン (I2Cマスターボードとの通信開始)
 * 引数: I2Cマスターボードのデバイスファイル
 * 戻値: 正常時 I2Cバスポインタ
 *       異常時 エラーが発生した場合はNULL (errnoで判断)
 */
I2CBus* i2cBusOpen(char* fileName) {
	I2CBus* bus;
	int error;

	/* バスリストロック */
	pthread_mutex_lock(&busses->mutex);
	/* すでに開かれているファイルか確認 */
	for(bus = busses->next; bus != busses; bus = bus->next ){
		if(strcmp(bus->name,fileName) == 0){
			ERROR_PRINT("%s: I2C bus has been already opend\n",fileName);
			pthread_mutex_unlock(&busses->mutex);
			errno = EIO;
			return NULL;
		}
	}
	/* バスリストクローズ */
	pthread_mutex_unlock(&busses->mutex);

	do{
		/* I2Bバスのメモリ確保 */
		bus = NULL;
		bus = (I2CBus*)malloc(sizeof(I2CBus));
		if (bus == NULL) {
			error = errno; /* errno退避 */
			ERROR_PRINT("malloc: %s\n",strerror(errno));
			break;
		}
		/* メンバ初期化 */
		bus->name = NULL;
		bus->devices = NULL;
		bus->fd = -1;
		pthread_mutex_init(&bus->mutex, NULL);

		/* ファイル名のメモリ確保 */
		bus->name = (char*)malloc(I2C_DEVICE_FILE_NAME_MAX+1);
		if (bus->name == NULL) {
			error = errno; /* errno退避 */
			ERROR_PRINT("malloc: %s\n",strerror(errno));
			break;
		}
		/* ファイル名保存 */
		strcpy(bus->name,fileName);

		/* I2Cデバイスリスト用のメモリ確保 */
		bus->devices = (I2CDevice*)malloc(sizeof(I2CDevice));
		if (bus->devices == NULL) {
			error = errno; /* errno退避 */
			ERROR_PRINT("malloc: %s\n",strerror(errno));
			break;
		}
		/* I2Cデバイスリスト初期化 */
		bus->devices->next = bus->devices;
		pthread_mutex_init(&bus->devices->mutex,NULL);

		/* I2Cマスターボードのデバイスファイルオープン */
		bus->fd = open(bus->name, O_RDWR);
		if (bus->fd < 0) {
			error = errno; /* errno退避 */
			ERROR_PRINT("open: %s: %s\n",bus->name,strerror(errno));
			break;
		}
		/* 通信設定変更 */
		error = i2cConfigTermios(bus);
		if (error != 0) {
			error = errno; /* errno退避 */
			break;
		}
		/* ブレーク信号送信 (I2Cマスターボードリセット) */
		error = tcsendbreak(bus->fd, 0);
		if (error != 0) {
			error = errno; /* errno退避 */
			ERROR_PRINT("tcsendbreak: %s: %s\n", bus->name, strerror(errno));
			break;
		}
		/* PICの再起動を待つ */
		usleep(100 * 1000);

		/* バスリストロック */
		pthread_mutex_lock(&busses->mutex);
		/* バスリスト更新 */
		bus->next = busses->next;
		busses->next = bus;
		/* バスリストアンロック */
		pthread_mutex_unlock(&busses->mutex);

		return bus;
	}while(0);

	/* エラー時の処理 */
	if(bus != NULL){
		pthread_mutex_destroy(&bus->mutex);
		if(bus->fd >= 0){
			error = close(bus->fd);
		}
		free(bus->devices);
		free(bus->name);
		free(bus);
	}
	/* errno復元 */
	errno = error;
	return NULL;
}

/*
 * バスのクローズ (I2Cマスターボードとの通信終了)
 * 引数: 通信終了するI2Cバスポインタ
 */
void i2cBusClose(I2CBus* bus) {
	I2CBus* prev;
	I2CDevice* dev;

	/* 引数確認 */
	if(bus == NULL){
		return;
	}

	/* バスロック */
	pthread_mutex_lock(&bus->mutex);

	/* 全デバイスの接続を切る */
	for(dev = bus->devices->next; dev != bus->devices; dev = dev->next){
		/* デバイスロック */
		pthread_mutex_lock(&dev->mutex);
		/* バスとの接続を切断 */
		dev->bus = NULL;
		/* デバイスアンロック */
		pthread_mutex_unlock(&dev->mutex);
	}

	/* バスリストロック */
	pthread_mutex_lock(&busses->mutex);
	/* バスリストから削除 */
	for(prev = bus->next; prev->next != bus; prev = prev->next);
	prev->next = bus->next;
	/* バスリストアンロック */
	pthread_mutex_unlock(&busses->mutex);

	/* まともなファイルディスクリプタか判定 */
	if(bus->fd >= 0){
		/* 送受信バッファフラッシュ */
		tcflush(bus->fd, TCIOFLUSH);
		/* I2Cマスターボードリセット */
		tcsendbreak(bus->fd, 0);
		/* ファイルクローズ */
		close(bus->fd);
	}
	/* バスアンロック */
	pthread_mutex_unlock(&bus->mutex);
	/* バスのmutex削除 */
	pthread_mutex_destroy(&bus->mutex);

	/* メモリ解放 */
	free(bus->devices);
	free(bus->name);
	free(bus);
}

/*
 * バスのリセット (ブレーク信号送信)
 * 戻値: 正常時 0
 *       異常時 負の数 (最上部の一覧を参照)
 */
int i2cBusReset(I2CBus* bus) {
	int error;
	/* 引数確認 */
	if (bus == NULL) {
		return I2C_INVAL;
	}
	/* バスロック */
	pthread_mutex_lock(&bus->mutex);
	error = 0;
	do{
		/* デバイスファイルを閉じる */
		close(bus->fd);
		/* もう一度開き直す */
		bus->fd = open(bus->name, O_RDWR);
		if (bus->fd < 0) { /* ファイルオープン失敗? */
			error = I2C_BUS_ERROR;
			break;
		}
		/* ブレーク信号送信 (I2Cマスターボードリセット) */
		error = tcsendbreak(bus->fd, 0);
		if (error != 0) {
			error = I2C_BUS_ERROR;
			break;
		}
		/* PICの再起動を待つ */
		usleep(100 * 1000);
	}while(0);
	/* バスアンロック */
	pthread_mutex_unlock(&bus->mutex);
	return error;
}

/*
 * デバイスが存在するか判定
 * 引数: デバイスが接続されているバス, デバイスのアドレス
 * 戻値: 存在する   1
 *       存在しない 0
 */
int i2cBusHasDevice(I2CBus* bus, uint8_t address){
	int result;
	int busHasDevice;
	uint8_t header[2];
	uint8_t response;

	/* 引数確認 */
	if (bus == NULL) {
		return 0;
	}
	/* アドレスが適切か確認 */
	if (address > I2C_ADDRESS_MAX) {
		return 0;
	}
	/* バスロック */
	pthread_mutex_lock(&bus->mutex);
	busHasDevice = 0;
	do{
		/* アドレスとデータサイズを送信 */
		header[0] = address << 1;
		header[1] = 0; /* 0Byte送信 */
		result = write(bus->fd, header, sizeof(header));
		if (result < sizeof(header)) {
			i2cBusReset(bus); /* 送信失敗時はバスをリセット */
			break;
		}
		/* スレーブの応答を確認 */
		result = read(bus->fd, &response, 1);
		if (result <= 0) {
			i2cBusReset(bus); /* 送信失敗時はバスをリセット */
			break;
		}
		switch (response) {
		case RESPONSE_SUCCESS: /* 正常応答 */
			busHasDevice = 1;
			break;
		case RESPONSE_SLAVE_NO_RESPONSE: /* 応答無し */
			break;
		default:  /* バス関連のエラー */
			i2cBusReset(bus);
			break;
		}
	}while(0);
	/* バスアンロック */
	pthread_mutex_unlock(&bus->mutex);
	/* 結果送信 */
	return busHasDevice;
}

/*
 * デバイスのオープン
 * 引数: デバイスが接続されているバス, デバイスのアドレス
 * 戻値: 正常時 I2Cデバイスポインタ
 *       異常時 NULL (errnoで判断)
 */
I2CDevice* i2cDeviceOpen(I2CBus* bus, uint8_t address) {
	I2CDevice* dev;
	int result;
	int error;
	uint8_t header[2];
	uint8_t response;
	/* 引数確認 */
	if (bus == NULL) {
		errno = EINVAL;
		return NULL;
	}
	/* アドレスが適切か確認 */
	if (address > I2C_ADDRESS_MAX) {
		ERROR_PRINT("I2C device address is too big\n");
		errno = EINVAL;
		return NULL;
	}

	/* バスロック */
	pthread_mutex_lock(&bus->mutex);

	/* すでに開かれているデバイスか確認 */
	for(dev = bus->devices->next; dev != bus->devices; dev = dev->next){
		if(dev->address == address){
			ERROR_PRINT("%s - 0x%02X: I2C device has been already opend\n",
					bus->name,address);
			pthread_mutex_unlock(&bus->devices->mutex);
			errno = EIO;
			return NULL;
		}
	}

	dev = NULL;
	do{
		/* メモリ確保 */
		dev = (I2CDevice*) malloc(sizeof(*dev));
		if (dev == NULL) { /* メモリ確保失敗 */
			error = errno; /* errno退避 */
			ERROR_PRINT("malloc: %s\n",strerror(errno));
			break;
		}
		/* メンバの初期化 */
		dev->bus = bus;
		dev->address = address;
		pthread_mutex_init(&dev->mutex,NULL);

		/* アドレスとデータサイズを送信して応答を受信 */
		header[0] = dev->address << 1;
		header[1] = 0;
		result = write(dev->bus->fd, header, sizeof(header));
		if (result < 0) {
			error = errno; /* errno退避 */
			ERROR_PRINT("write: %s: %s\n", bus->name, strerror(errno));
			pthread_mutex_unlock(&bus->mutex);
			break;
		} else if (result < sizeof(header)) {
			error = EIO;
			ERROR_PRINT("I2C bus error\n");
			pthread_mutex_unlock(&bus->mutex);
			break;
		}
		result = read(dev->bus->fd, &response, 1);

		/* スレーブの応答を確認 */
		if (result < 0) {
			error = errno; /* errno退避 */
			ERROR_PRINT("read: %s: %s\n", bus->name, strerror(errno));
			break;
		} else if (result == 0) {
			error = EIO;
			ERROR_PRINT("I2C bus give no response (time-out)\n");
			break;
		}
		switch (response) {
		case RESPONSE_SUCCESS:
			error = 0;
			break;
		case RESPONSE_SLAVE_NO_RESPONSE:
			error = -1;
			ERROR_PRINT("I2C device give no response\n");
			break;
		case RESPONSE_RS232C_ERROR:
			error = -1;
			ERROR_PRINT("RS232C error\n");
			break;
		default:
			error = -1;
			ERROR_PRINT("I2C bus unknown error\n");
			break;
		}
		if(error){ error = EIO; break; }

		/* デバイスリストに追加 */
		dev->next = bus->devices->next;
		bus->devices->next = dev;
		/* バスアンロック */
		pthread_mutex_unlock(&bus->mutex);

		return dev;
	}while(0);

	/* エラー時の処理 */
	pthread_mutex_unlock(&bus->mutex);
	free(dev);
	errno = error;
	return NULL;
}

/*
 * デバイスのクローズ
 * 引数: 通信終了するI2Cデバイスポインタ
 */
void i2cDeviceClose(I2CDevice* dev) {
	I2CDevice* prev;
	/* 引数確認 */
	if (dev == NULL) {
		return;
	}
	/* バスがロックできるまでループ */
	while(1){
		/* デバイスロック */
		pthread_mutex_lock(&dev->mutex);
		/* バスがもう閉じられていたらメモリ解放して終了 */
		if(dev->bus == NULL){
			/* デバイスアンロック */
			pthread_mutex_unlock(&dev->mutex);
			/* メモリ解放 */
			free(dev);
			return;
		}
		/* バスがロックできたらループ終了 */
		if(pthread_mutex_trylock(&dev->bus->mutex) != EBUSY){
			break;
		}
		/* 無理なら最初からやり直し */
		pthread_mutex_unlock(&dev->mutex);
	}
	/* デバイスリストからデバイスを削除 */
	for(prev = dev->next; prev->next != dev; prev = prev->next);
	prev->next = dev->next;
	/* バスアンロック */
	pthread_mutex_unlock(&dev->bus->mutex);
	/* デバイスアンロック */
	pthread_mutex_unlock(&dev->mutex);
	/* メモリ解放 */
	free(dev);
	return;
}

/*
 * デバイスにデータ送信
 * 引数: I2Cデバイスポインタ, 書き込むデータ, データサイズ
 * 戻値: 正常時 送信できたバイト数
 *       異常時 負の数 (最上部の一覧を参照)
 */
int i2cWrite(I2CDevice* dev, void* buf, uint8_t size) {
	int result;
	uint8_t header[2];
	uint8_t response;

	/* 引数確認 */
	if (dev == NULL) {
		result = I2C_INVAL;
		return result;
	}

	/* デバイスロック */
	pthread_mutex_lock(&dev->mutex);
	/* バスがまだ生きているか確認 */
	if(dev->bus == NULL){
		pthread_mutex_unlock(&dev->mutex);
		ERROR_PRINT("I2C bus has been already closed\n");
		result = I2C_BUS_ERROR;
		return result;
	}

	/* バスロック */
	pthread_mutex_lock(&dev->bus->mutex);
	do{
		/* アドレスとデータサイズを送信 */
		header[0] = dev->address << 1;
		header[1] = size;
		result = write(dev->bus->fd, header, sizeof(header));
		if (result < 0) {
			ERROR_PRINT("write: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result < sizeof(header)) {
			ERROR_PRINT("I2C bus error\n");
			result = I2C_BUS_ERROR;
			break;
		}
		/* スレーブの応答を確認 */
		result = read(dev->bus->fd, &response, 1);
		if (result < 0) {
			ERROR_PRINT("read: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result == 0) {
			ERROR_PRINT("read: I2C bus give no response (time-out)\n");
			result = I2C_BUS_ERROR;
			break;
		}
		switch (response) {
		case RESPONSE_SUCCESS:
			result = 0;
			break;
		case RESPONSE_SLAVE_NO_RESPONSE:
			ERROR_PRINT("I2C device give no response\n");
			result = I2C_DEVICE_ERROR;
			break;
		case RESPONSE_RS232C_ERROR:
			ERROR_PRINT("RS232C error\n");
			result = I2C_BUS_ERROR;
			break;
		default:
			ERROR_PRINT( "I2C bus unknown error\n");
			result = I2C_BUS_ERROR;
			break;
		}
		if(result != 0){ break; }

		/* データを送信する */
		result = write(dev->bus->fd, buf, size);
		if (result < 0) {
			ERROR_PRINT("write: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result < size) {
			ERROR_PRINT("I2C bus error\n");
			result = I2C_BUS_ERROR;
			break;
		}
		/* 結果を受信する */
		result = read(dev->bus->fd, &response, 1);
		if (result < 0) {
			ERROR_PRINT("read: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result == 0) {
			ERROR_PRINT("read: I2C bus give no response (time-out)\n");
			result = I2C_BUS_ERROR;
			break;
		}
		result = response;

		/* 一応エラー出力 */
		if (result < size) {
			ERROR_PRINT( "I2C device recieved only %d Byte of %d Byte\n",
					(int)result,(int)size);
		}
	}while(0);
	/* バスアンロック */
	pthread_mutex_unlock(&dev->bus->mutex);
	/* デバイスアンロック */
	pthread_mutex_unlock(&dev->mutex);

	/* バスにエラーが発生した場合はリセット */
	if(result == I2C_BUS_ERROR){
		i2cBusReset(dev->bus);
	}
	return result;
}

/*
 * デバイスからデータ受信
 * 引数: I2Cデバイスポインタ, 受信バッファ, データサイズ
 * 戻値: 正常時 受信できたバイト数
 *       異常時 負の数 (最上部の一覧を参照)
 */
int i2cRead(I2CDevice* dev, void* buf, uint8_t size) {
	int i;
	int result;
	uint8_t header[2];
	uint8_t response;

	/* 引数確認 */
	if (dev == NULL) {
		result = I2C_INVAL;
		return result;
	}

	/* デバイスロック */
	pthread_mutex_lock(&dev->mutex);
	/* バスがまだ生きているか確認 */
	if(dev->bus == NULL){
		pthread_mutex_unlock(&dev->mutex);
		ERROR_PRINT("I2C bus has been already closed\n");
		result = I2C_BUS_ERROR;
		return result;
	}

	/* バスロック */
	pthread_mutex_lock(&dev->bus->mutex);
	do{
		/* アドレスとデータサイズを送信 */
		header[0] = dev->address << 1 | 0x01;
		header[1] = size;
		result = write(dev->bus->fd, header, sizeof(header));
		if (result < 0) {
			ERROR_PRINT("write: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result < sizeof(header)) {
			ERROR_PRINT("I2C bus error\n");
			result = I2C_BUS_ERROR;
			break;
		}
		/* スレーブの応答を確認 */
		result = read(dev->bus->fd, &response, 1);
		if (result < 0) {
			ERROR_PRINT("read: %s\n",strerror(errno));
			result = I2C_BUS_ERROR;
			break;
		} else if (result == 0) {
			ERROR_PRINT("read: I2C bus give no response (time-out)\n");
			result = I2C_BUS_ERROR;
			break;
		}
		switch (response) {
		case RESPONSE_SUCCESS:
			result = 0;
			break;
		case RESPONSE_SLAVE_NO_RESPONSE:
			ERROR_PRINT("I2C device give no response\n");
			result = I2C_DEVICE_ERROR;
			break;
		case RESPONSE_RS232C_ERROR:
			ERROR_PRINT("RS232C error\n");
			result = I2C_BUS_ERROR;
			break;
		default:
			ERROR_PRINT( "I2C bus unknown error\n");
			result = I2C_BUS_ERROR;
			break;
		}
		if(result != 0){ break; }

		/* 続きを受信する */
		for (i = 0; i < size; i += result) {
			result = read(dev->bus->fd, ((uint8_t*) buf) + i, size - i);
			if (result < 0) {
				ERROR_PRINT("read: %s\n",strerror(errno));
				result = I2C_BUS_ERROR;
				break;
			} else if (result == 0) {
				ERROR_PRINT("read: I2C bus give no response (time-out) at %d/%d\n",i,size);
				result = I2C_BUS_ERROR;
				break;
			}
		}
		if(i == size){
			result = size;
		}
	}while(0);
	/* バスアンロック */
	pthread_mutex_unlock(&dev->bus->mutex);
	/* デバイスアンロック */
	pthread_mutex_unlock(&dev->mutex);

	/* バスにエラーが発生した場合はリセット */
	if(result == I2C_BUS_ERROR){
		i2cBusReset(dev->bus);
	}
	return result;
}
